﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Analysis;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Persistence.Default.Xml;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using RunCreationClient = HeuristicLab.Clients.OKB.RunCreation.RunCreationClient;
using SingleObjectiveOKBProblem = HeuristicLab.Clients.OKB.RunCreation.SingleObjectiveOKBProblem;

namespace HeuristicLab.OptimizationExpertSystem.Common {
  [Item("Expert-System", "Currently in experimental phase, an expert system that makes algorithm suggestions based on fitness landscape analysis features and an optimization knowledge base.")]
  [StorableClass]
  [Creatable(CreatableAttribute.Categories.TestingAndAnalysis, Priority = 119)]
  public sealed class ExpertSystem : NamedItem, IStorableContent, INotifyPropertyChanged {

    public string Filename { get; set; }

    public static new Image StaticItemImage {
      get { return VSImageLibrary.Library; }
    }

    [Storable]
    private int maximumEvaluations;
    public int MaximumEvaluations {
      get { return maximumEvaluations; }
      set {
        if (maximumEvaluations == value) return;
        maximumEvaluations = value;
        OnPropertyChanged("MaximumEvaluations");
        UpdateSuggestions();
      }
    }

    [Storable]
    private RunCollection runs;
    public RunCollection Runs {
      get { return runs; }
    }

    [Storable]
    private RunCollection knowledgeBase;
    public RunCollection KnowledgeBase {
      get { return knowledgeBase; }
      set {
        if (knowledgeBase == value) return;
        knowledgeBase = value;
        OnPropertyChanged("KnowledgeBase");
      }
    }

    [Storable]
    private SingleObjectiveOKBProblem problem;
    public SingleObjectiveOKBProblem Problem {
      get { return problem; }
      set {
        if (problem == value) return;
        problem = value;
        OnPropertyChanged("Problem");
        UpdateSuggestions();
      }
    }

    [Storable]
    private ItemList<IAlgorithm> suggestedInstances;
    private ReadOnlyItemList<IAlgorithm> readOnlySuggestedInstances;
    public ReadOnlyItemList<IAlgorithm> SuggestedInstances {
      get { return readOnlySuggestedInstances; }
    }

    [Storable]
    private RunCollection problemInstances;
    public RunCollection ProblemInstances {
      get { return problemInstances; }
      set {
        if (problemInstances == value) return;
        problemInstances = value;
        OnPropertyChanged("ProblemInstances");
      }
    }

    [Storable]
    private ResultCollection currentResult;
    public ResultCollection CurrentResult {
      get { return currentResult; }
      set {
        if (currentResult == value) return;
        currentResult = value;
        OnPropertyChanged("CurrentResult");
      }
    }

    [Storable(Name = "newSolutionPool")]
    private CheckedItemList<IScope> solutionPool;
    public CheckedItemList<IScope> SolutionPool {
      get { return solutionPool; }
      set {
        if (solutionPool == value) return;
        solutionPool = value;
        OnPropertyChanged("SolutionPool");
      }
    }
    
    [Storable]
    private BidirectionalLookup<long, IRun> algorithmId2RunMapping;
    [Storable]
    private BidirectionalDictionary<long, IAlgorithm> algorithmId2AlgorithmInstanceMapping;

    [Storable]
    private Run currentInstance;

    private bool Maximization {
      get {
        return Problem != null && Problem.ProblemId >= 0 && ((IValueParameter<BoolValue>)Problem.MaximizationParameter).Value.Value;
      }
    }

    [StorableConstructor]
    private ExpertSystem(bool deserializing) : base(deserializing) { }
    private ExpertSystem(ExpertSystem original, Cloner cloner)
      : base(original, cloner) {
      runs = cloner.Clone(original.runs);
      knowledgeBase = cloner.Clone(original.knowledgeBase);
      suggestedInstances = cloner.Clone(original.suggestedInstances);
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      problemInstances = cloner.Clone(original.problemInstances);
      problem = cloner.Clone(original.problem);
      algorithmId2RunMapping = new BidirectionalLookup<long, IRun>();
      foreach (var kvp in original.algorithmId2RunMapping.FirstEnumerable) {
        algorithmId2RunMapping.AddRangeFirst(kvp.Key, kvp.Select(cloner.Clone));
      }
      algorithmId2AlgorithmInstanceMapping = new BidirectionalDictionary<long, IAlgorithm>();
      foreach (var kvp in original.algorithmId2AlgorithmInstanceMapping) {
        algorithmId2AlgorithmInstanceMapping.Add(kvp.Key, cloner.Clone(kvp.Value));
      }
      currentResult = cloner.Clone(original.currentResult);
      solutionPool = cloner.Clone(original.solutionPool);
      currentInstance = cloner.Clone(original.currentInstance);
      RegisterEventHandlers();
    }
    public ExpertSystem() {
      Name = ItemName;
      Description = ItemDescription;
      runs = new RunCollection();
      knowledgeBase = new RunCollection();
      suggestedInstances = new ItemList<IAlgorithm>();
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      problemInstances = new RunCollection();
      problem = new SingleObjectiveOKBProblem();
      solutionPool = new CheckedItemList<IScope>();
      algorithmId2RunMapping = new BidirectionalLookup<long, IRun>();
      algorithmId2AlgorithmInstanceMapping = new BidirectionalDictionary<long, IAlgorithm>();
      RegisterEventHandlers();
    }

    private void ProblemOnProblemChanged(object sender, EventArgs eventArgs) {
      if (Problem == null) return;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new ExpertSystem(this, cloner);
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      RegisterEventHandlers();
    }

    private void RegisterEventHandlers() {
      problem.ProblemChanged += ProblemOnProblemChanged;
      runs.CollectionReset += InformationChanged;
      runs.ItemsAdded += InformationChanged;
      runs.ItemsRemoved += InformationChanged;
      runs.Reset += InformationChanged;
      runs.UpdateOfRunsInProgressChanged += InformationChanged;
      knowledgeBase.CollectionReset += InformationChanged;
      knowledgeBase.ItemsAdded += InformationChanged;
      knowledgeBase.ItemsRemoved += InformationChanged;
    }

    private void InformationChanged(object sender, EventArgs e) {
      var runCollection = sender as RunCollection;
      if (runCollection != null && runCollection.UpdateOfRunsInProgress) return;
      UpdateSuggestions();
    }

    public void UpdateInstanceProjection() {
      currentInstance = null;
      var filter = new HashSet<string>();

      if (Problem.ProblemId != -1) {
        currentInstance = new Run { Name = Problem.Name };
        Problem.Problem.CollectParameterValues(currentInstance.Parameters);
        foreach (var c in RunCreationClient.Instance.GetCharacteristicValues(Problem.ProblemId)) {
          var key = "Characteristic." + c.Name;
          currentInstance.Results.Add(key, RunCreationClient.Instance.ConvertToItem(c));
          filter.Add(key);
        }
      }
      // TODO: There is a problem with missing values that has to be solved
      // The common set of characteristics among all problems may be too small to be useful
      // It has to be decided somehow which problem instances to include in order to obtain the set of features that is expressive and available
      var allCharacteristics = ProblemInstances.Select(x => new HashSet<string>(x.Results.Where(y => y.Key.StartsWith("Characteristic.", StringComparison.Ordinal)).Select(y => y.Key))).ToList();
      var commonCharacteristics = filter.Count > 0 ? filter : allCharacteristics[0];
      for (var i = 0; i < allCharacteristics.Count; i++)
        commonCharacteristics.IntersectWith(allCharacteristics[i]);

      if (commonCharacteristics.Count == 0) return;

      var instances = new Dictionary<string, double[]>();
      foreach (var i in ProblemInstances)
        instances[i.Name] = i.Results.Where(x => commonCharacteristics.Contains(x.Key)).OrderBy(x => x.Key).Select(x => (double)((dynamic)x.Value).Value).ToArray();
      if (currentInstance != null)
        instances[currentInstance.Name] = currentInstance.Results.Where(x => commonCharacteristics.Contains(x.Key)).OrderBy(x => x.Key).Select(x => (double)((dynamic)x.Value).Value).ToArray();

      var allValues = instances.SelectMany(x => x.Value).ToList();
      var avg = allValues.Average();
      var stdev = allValues.StandardDeviation();

      // normalize characteristic values by transforming them to their z-score
      foreach (var key in instances.Keys.ToList()) {
        var arr = instances[key];
        for (var i = 0; i < arr.Length; i++) {
          arr[i] = (arr[i] - avg) / stdev;
        }
      }

      var key2Idx = new BidirectionalDictionary<string, int>();
      foreach (var kvp in instances.Select((k, i) => new { Index = i, Key = k.Key }))
        key2Idx.Add(kvp.Key, kvp.Index);

      #region MDS
      Func<double[], double[], double> euclid = (a, b) => Math.Sqrt(a.Zip(b, (x, y) => (x - y)).Sum(x => x * x));
      var num = instances.Count;
      var matrix = new DoubleMatrix(num, num);
      for (var i = 0; i < num - 1; i++) {
        for (var j = i + 1; j < num; j++) {
          matrix[i, j] = matrix[j, i] = euclid(instances[key2Idx.GetBySecond(i)], instances[key2Idx.GetBySecond(j)]);
        }
      }

      var coords = MultidimensionalScaling.KruskalShepard(matrix);
      #endregion
      #region PCA
      var ds = new double[instances.Count, commonCharacteristics.Count];
      foreach (var instance in instances) {
        var arr = instance.Value;
        for (var feature = 0; feature < arr.Length; feature++)
          ds[key2Idx.GetByFirst(instance.Key), feature] = arr[feature];
      }

      int info;
      double[] s2;
      double[,] v;
      alglib.pcabuildbasis(ds, ds.GetLength(0), ds.GetLength(1), out info, out s2, out v);
      #endregion

      ProblemInstances.UpdateOfRunsInProgress = true;
      try {
        foreach (var instance in ProblemInstances) {
          double x = 0, y = 0;
          for (var feature = 0; feature < ds.GetLength(1); feature++) {
            x += ds[key2Idx.GetByFirst(instance.Name), feature] * v[feature, 0];
            y += ds[key2Idx.GetByFirst(instance.Name), feature] * v[feature, 1];
          }

          IItem item;
          if (instance.Results.TryGetValue("Projection.PCA.X", out item)) {
            ((DoubleValue)item).Value = x;
          } else instance.Results.Add("Projection.PCA.X", new DoubleValue(x));
          if (instance.Results.TryGetValue("Projection.PCA.Y", out item)) {
            ((DoubleValue)item).Value = y;
          } else instance.Results.Add("Projection.PCA.Y", new DoubleValue(y));

          if (instance.Results.TryGetValue("Projection.MDS.X", out item)) {
            ((DoubleValue)item).Value = coords[key2Idx.GetByFirst(instance.Name), 0];
          } else instance.Results.Add("Projection.MDS.X", new DoubleValue(coords[key2Idx.GetByFirst(instance.Name), 0]));
          if (instance.Results.TryGetValue("Projection.MDS.Y", out item)) {
            ((DoubleValue)item).Value = coords[key2Idx.GetByFirst(instance.Name), 1];
          } else instance.Results.Add("Projection.MDS.Y", new DoubleValue(coords[key2Idx.GetByFirst(instance.Name), 1]));
        }
      } finally { ProblemInstances.UpdateOfRunsInProgress = false; }

      if (currentInstance != null) {
        double x = 0, y = 0;
        for (var feature = 0; feature < ds.GetLength(1); feature++) {
          x += ds[key2Idx.GetByFirst(currentInstance.Name), feature] * v[feature, 0];
          y += ds[key2Idx.GetByFirst(currentInstance.Name), feature] * v[feature, 1];
        }

        IItem item;
        if (currentInstance.Results.TryGetValue("Projection.PCA.X", out item)) {
          ((DoubleValue)item).Value = x;
        } else currentInstance.Results.Add("Projection.PCA.X", new DoubleValue(x));
        if (currentInstance.Results.TryGetValue("Projection.PCA.Y", out item)) {
          ((DoubleValue)item).Value = y;
        } else currentInstance.Results.Add("Projection.PCA.Y", new DoubleValue(y));

        if (currentInstance.Results.TryGetValue("Projection.MDS.X", out item)) {
          ((DoubleValue)item).Value = coords[key2Idx.GetByFirst(currentInstance.Name), 0];
        } else currentInstance.Results.Add("Projection.MDS.X", new DoubleValue(coords[key2Idx.GetByFirst(currentInstance.Name), 0]));
        if (currentInstance.Results.TryGetValue("Projection.MDS.Y", out item)) {
          ((DoubleValue)item).Value = coords[key2Idx.GetByFirst(currentInstance.Name), 1];
        } else currentInstance.Results.Add("Projection.MDS.Y", new DoubleValue(coords[key2Idx.GetByFirst(currentInstance.Name), 1]));
      }
    }

    public Tuple<double, double> ProjectCurrentInstance(string projection) {
      if (currentInstance == null) return null;
      var xKey = "Projection." + projection + ".X";
      var yKey = "Projection." + projection + ".Y";
      if (!currentInstance.Results.ContainsKey(xKey) || !currentInstance.Results.ContainsKey(yKey))
        return null;
      var x = ((DoubleValue)currentInstance.Results[xKey]).Value;
      var y = ((DoubleValue)currentInstance.Results[yKey]).Value;
      return Tuple.Create(x, y);
    }

    private static readonly HashSet<string> InterestingValueNames = new HashSet<string>() {
      "QualityPerEvaluations", "Problem Name", "Problem Type", "Algorithm Name", "Algorithm Type", "Maximization", "BestKnownQuality"
    };

    public void StartAlgorithmAsync(int index) {
      var selectedInstance = suggestedInstances[index];
      var algorithm = (IAlgorithm)selectedInstance.Clone();
      algorithm.Problem = Problem.CloneProblem();
      algorithm.Prepare(true);
      IParameter stopParam;
      var monitorStop = true;
      if (algorithm.Parameters.TryGetValue("MaximumEvaluations", out stopParam)) {
        var maxEvalParam = stopParam as IValueParameter<Data.IntValue>;
        if (maxEvalParam != null) {
          maxEvalParam.Value.Value = MaximumEvaluations;
          monitorStop = false;
        }
      }
      algorithm.ExecutionStateChanged += AlgorithmOnExecutionStateChanged;
      algorithm.ExceptionOccurred += AlgorithmOnExceptionOccurred;
      if (monitorStop) algorithm.ExecutionTimeChanged += AlgorithmOnExecutionTimeChanged;

      algorithm.Start();
    }

    public void StartAlgorithm(int index) {
      var selectedInstance = suggestedInstances[index];
      var algorithm = (IAlgorithm)selectedInstance.Clone();
      algorithm.Problem = Problem.CloneProblem();
      algorithm.Prepare(true);
      IParameter stopParam;
      var monitorStop = true;
      if (algorithm.Parameters.TryGetValue("MaximumEvaluations", out stopParam)) {
        var maxEvalParam = stopParam as IValueParameter<Data.IntValue>;
        if (maxEvalParam != null) {
          maxEvalParam.Value.Value = MaximumEvaluations;
          monitorStop = false;
        }
      }
      algorithm.ExecutionStateChanged += AlgorithmOnExecutionStateChanged;
      algorithm.ExceptionOccurred += AlgorithmOnExceptionOccurred;
      if (monitorStop) algorithm.ExecutionTimeChanged += AlgorithmOnExecutionTimeChanged;

      using (algWh = new AutoResetEvent(false)) {
        algorithm.Start();
        algWh.WaitOne();
      }
      algWh = null;
    }

    private AutoResetEvent algWh;

    private void AlgorithmOnExecutionStateChanged(object sender, EventArgs eventArgs) {
      var alg = sender as IAlgorithm;
      if (alg == null) return;
      if (alg.ExecutionState == ExecutionState.Started) {
        CurrentResult = alg.Results;
      } else if (alg.ExecutionState == ExecutionState.Stopped) {
        alg.ExecutionStateChanged -= AlgorithmOnExecutionStateChanged;
        alg.ExceptionOccurred -= AlgorithmOnExceptionOccurred;
        alg.ExecutionTimeChanged -= AlgorithmOnExecutionTimeChanged;
        foreach (var solution in alg.Results.Where(x => x.Name.ToLower().Contains("solution")).Select(x => x.Value).OfType<IScope>()) {
          solutionPool.Add(solution);
        }
        Runs.Add(alg.Runs.Last());
        if (algWh != null) algWh.Set();
      }
    }

    private void AlgorithmOnExceptionOccurred(object sender, EventArgs<Exception> eventArgs) {
      var alg = sender as IAlgorithm;
      if (alg == null) return;
      alg.ExecutionStateChanged -= AlgorithmOnExecutionStateChanged;
      alg.ExceptionOccurred -= AlgorithmOnExceptionOccurred;
      alg.ExecutionTimeChanged -= AlgorithmOnExecutionTimeChanged;
      if (algWh != null) algWh.Set();
    }

    private void AlgorithmOnExecutionTimeChanged(object sender, EventArgs eventArgs) {
      var alg = sender as IAlgorithm;
      if (alg == null) return;
      IResult evalSolResult;
      if (!alg.Results.TryGetValue("EvaluatedSolutions", out evalSolResult) || !(evalSolResult.Value is Data.IntValue)) return;
      var evalSols = ((Data.IntValue)evalSolResult.Value).Value;
      if (evalSols >= MaximumEvaluations) alg.Stop();
    }

    public async void UpdateKnowledgeBaseAsync(IProgress progress) {
      progress.Start("Updating Knowledge Base from OKB");
      await Task.Factory.StartNew(() => { DoUpdateKnowledgeBase(progress); }, TaskCreationOptions.LongRunning);
    }

    public void UpdateKnowledgeBase() {
      DoUpdateKnowledgeBase(new Progress(string.Empty, ProgressState.Started));
    }

    private void DoUpdateKnowledgeBase(IProgress progress) {
      var queryClient = Clients.OKB.Query.QueryClient.Instance;
      var adminClient = Clients.OKB.Administration.AdministrationClient.Instance;
      try {
        progress.Status = "Downloading run information...";
        progress.ProgressValue = 0;
        // FIXME: How to tell if refresh is necessary?
        queryClient.Refresh();
        progress.ProgressValue = 0.5;
        progress.Status = "Downloading algorithm and problem instance information...";
        // FIXME: How to tell if refresh is necessary?
        adminClient.Refresh();

        var probInstance = adminClient.Problems.SingleOrDefault(x => x.Id == Problem.ProblemId);
        if (probInstance == null) throw new InvalidOperationException("The chosen problem instance cannot be found in the OKB.");
        var probClassId = probInstance.ProblemClassId;

        var problemClassFilter = (Clients.OKB.Query.StringComparisonAvailableValuesFilter)queryClient.Filters.Single(x => x.Label == "Problem Class Name");
        problemClassFilter.Value = adminClient.ProblemClasses.Single(x => x.Id == probClassId).Name;

        progress.Status = "Downloading problem instances...";
        progress.ProgressValue = 0;
        var p = 0;
        ProblemInstances.UpdateOfRunsInProgress = true;
        ProblemInstances.Clear();
        var totalProblems = adminClient.Problems.Count(x => x.ProblemClassId == probClassId);
        foreach (var problInst in adminClient.Problems.Where(x => x.ProblemClassId == probClassId)) {
          progress.Status = string.Format("Downloading problem {0} (okb-id: {1})....", problInst.Name, problInst.Id);
          var data = Clients.OKB.Administration.AdministrationClient.GetProblemData(problInst.Id);
          if (data != null) {
            using (var stream = new MemoryStream(data)) {
              try {
                var prob = (IProblem)XmlParser.Deserialize<IContent>(stream);
                var probRun = new Run() { Name = prob.Name };
                prob.CollectParameterValues(probRun.Parameters);
                probRun.Parameters["Problem Name"] = new StringValue(prob.Name);
                probRun.Parameters["Problem Type"] = new StringValue(prob.GetType().Name);
                progress.Status += Environment.NewLine + "Downloading characteristics...";
                foreach (var v in RunCreationClient.Instance.GetCharacteristicValues(problInst.Id)) {
                  probRun.Results.Add("Characteristic." + v.Name, RunCreationClient.Instance.ConvertToItem(v));
                }
                ProblemInstances.Add(probRun);
              } catch { }
              stream.Close();
            }
          }
          p++;
          progress.ProgressValue = p / (double)totalProblems;
        }

        algorithmId2AlgorithmInstanceMapping.Clear();
        progress.Status = "Downloading algorithm instances...";
        progress.ProgressValue = 0;
        p = 0;
        foreach (var algInst in adminClient.Algorithms) {
          progress.Status = string.Format("Downloading algorithm {0} (okb-id: {1})...", algInst.Name, algInst.Id);
          var data = Clients.OKB.Administration.AdministrationClient.GetAlgorithmData(algInst.Id);
          if (data != null) {
            using (var stream = new MemoryStream(data)) {
              try {
                var alg = (IAlgorithm)XmlParser.Deserialize<IContent>(stream);
                algorithmId2AlgorithmInstanceMapping.Add(algInst.Id, alg);
              } catch { }
              stream.Close();
            }
          }
          p++;
          progress.ProgressValue = p / (double)adminClient.Algorithms.Count;
        }

        var interestingValues = queryClient.ValueNames.Where(x => InterestingValueNames.Contains(x.Name)).ToList();

        progress.Status = "Obtaining number of runs...";
        progress.ProgressValue = 0;
        p = 0;
        var count = queryClient.GetNumberOfRuns(problemClassFilter);
        if (count == 0) return;
        
        var runIds = queryClient.GetRunIds(problemClassFilter).ToList();
        var conversions = new List<Task>();
        var runList = new List<IRun>();
        while (p < count) {
          var nextIds = runIds.Skip(p).Take(500).ToList();
          progress.Status = string.Format("Downloading runs {0} to {1} of {2}...", p, p + nextIds.Count, count);
          var okbRuns = queryClient.GetRunsWithValues(nextIds, true, interestingValues);
          conversions.Add(Task.Factory.StartNew(() => {
            var hlRuns = okbRuns.AsParallel().Select(x => new { AlgorithmId = x.Algorithm.Id, Run = queryClient.ConvertToOptimizationRun(x) }).ToList();
            lock (runList) {
              foreach (var r in hlRuns) {
                algorithmId2RunMapping.Add(r.AlgorithmId, r.Run);
                runList.Add(r.Run);
              }
            }
          }));
          p += nextIds.Count;
          progress.ProgressValue = p / (double)count;
        }
        Task.WaitAll(conversions.ToArray());

        progress.Status = "Finishing...";
        var algInstRunDict = runList.Where(x => x.Parameters.ContainsKey("Problem Name") && x.Parameters["Problem Name"] is StringValue)
                                          .GroupBy(x => ((StringValue)x.Parameters["Problem Name"]).Value)
                                          .ToDictionary(x => x.Key, x => x.GroupBy(y => ((StringValue)y.Parameters["Algorithm Name"]).Value)
                                                                                  .ToDictionary(y => y.Key, y => y.ToList()));

        foreach (var instance in ProblemInstances) {
          IItem probNameParam;
          if (!instance.Parameters.TryGetValue("Problem Name", out probNameParam)) continue;

          var probInstanceName = ((StringValue)probNameParam).Value;
          var maximization = ((BoolValue)instance.Parameters["Maximization"]).Value;
          
          IItem bkParam;
          if (!instance.Parameters.TryGetValue("BestKnownQuality", out bkParam) || !(bkParam is DoubleValue)) {
            Dictionary<string, List<IRun>> algRuns;
            if (!algInstRunDict.TryGetValue(probInstanceName, out algRuns)) continue;
            var list = algRuns.SelectMany(x => x.Value)
                              .Where(x => x.Results.ContainsKey("QualityPerEvaluations"))
                              .Select(x => ((IndexedDataTable<double>)x.Results["QualityPerEvaluations"]).Rows.First().Values.Last().Item2);
            bkParam = new DoubleValue(maximization ? list.Max() : list.Min());
            instance.Parameters["BestKnownQuality"] = bkParam;
          } else bkParam = instance.Parameters["BestKnownQuality"];

          var bkQuality = ((DoubleValue)bkParam).Value;

          if (!algInstRunDict.ContainsKey(probInstanceName)) continue;
          foreach (var kvp in algInstRunDict[probInstanceName]) {
            // TODO: Things needs to be configurable here (table name, targets)
            foreach (var target in new[] { 1, 1.01, 1.05, 1.1, 1.2, 1.5 }) {
              var result = ExpectedRuntimeHelper.CalculateErt(kvp.Value, "QualityPerEvaluations", bkQuality * target, maximization);
              var resultName = kvp.Key + "@" + ((target - 1) * 100) + "%";
              IItem item;
              if (instance.Results.TryGetValue(resultName, out item)) {
                ((DoubleValue)item).Value = Math.Log10(result.ExpectedRuntime);
              } else instance.Results.Add(resultName, new DoubleValue(Math.Log10(result.ExpectedRuntime)));
            }
          }
        }
        KnowledgeBase = new RunCollection(runList);
      } finally { progress.Finish(); ProblemInstances.UpdateOfRunsInProgress = false; }
    }

    private void UpdateSuggestions() {
      if (Problem == null) return;
      var instances = new SortedList<double, IAlgorithm>();
      foreach (var relevantRuns in knowledgeBase.GroupBy(x => algorithmId2RunMapping.GetBySecond(x).Single())) {
        var algorithm = algorithmId2AlgorithmInstanceMapping.GetByFirst(relevantRuns.Key);
        var avgQuality = 0.0;
        var counter = 0;
        foreach (var problemRuns in relevantRuns.GroupBy(x => ((StringValue)x.Parameters["Problem Name"]).Value)) {
          var probInstance = ProblemInstances.SingleOrDefault(x => x.Name == problemRuns.Key);
          if (probInstance == null) continue;
          var bkQuality = ((DoubleValue)probInstance.Parameters["BestKnownQuality"]).Value;
          foreach (var run in problemRuns) {
            var performanceGraph = ((IndexedDataTable<double>)run.Results["QualityPerEvaluations"]);
            try {
              avgQuality += performanceGraph.Rows.First().Values.TakeWhile(x => x.Item1 < MaximumEvaluations).Last().Item2 / bkQuality;
              counter++;
            } catch {
              continue;
            }
          }
        }
        avgQuality /= counter;
        instances.Add(avgQuality, algorithm);
      }

      suggestedInstances.Clear();
      var instanceLadder = instances.Select(x => (IAlgorithm)x.Value.Clone()).ToList();
      if (Maximization) instanceLadder.Reverse();
      suggestedInstances.AddRange(instanceLadder);
    }

    public event PropertyChangedEventHandler PropertyChanged;
    private void OnPropertyChanged(string propertyName) {
      var handler = PropertyChanged;
      if (handler != null) handler(this, new PropertyChangedEventArgs(propertyName));
    }
  }
}
